from __future__ import annotations

import pathlib
import urllib.parse
from typing import Any, Iterator, Sequence

from airflow.providers.google.suite.hooks.drive import GoogleDriveHook

from astro.constants import FileLocation
from astro.files.locations.base import BaseFileLocation


def _quote(s: str) -> str:
    s = s.replace("'", "\\'")  # Escape single quote characters.
    return f"'{s}'"  # Quote.


def _find_item_id(path_parts: str, *, conn: Any) -> str:
    """Drill down the directory hierarchy to find the ID of an item.

    Google Drive does not have a very good random access API by path, so we
    must look at each directory, level by level, to find the item we want.
    """
    current_parent = "root"
    folders = path_parts.split("/")
    # First tries to enter directories
    for current_folder in folders:
        conditions = [
            "mimeType = 'application/vnd.google-apps.folder'",
            f"name='{current_folder}'",
        ]
        if current_parent != "root":
            conditions.append(f"'{current_parent}' in parents")
        result = (
            conn.files().list(q=" and ".join(conditions), spaces="drive", fields="files(id, name)").execute()
        )
        files = result.get("files", [])
        if not files:
            # If the directory does not exist, break loops
            break
        current_parent = files[0].get("id")
    return current_parent


def _find_contains(folder_id: str, pattern: str, *, conn: Any) -> Iterator[str]:
    """Find files in a directory matching given pattern.

    This uses *contains* to search. See Google Drive documentation to
    understand the implications:
    https://developers.google.com/drive/api/guides/ref-search-terms
    """
    page_token = None
    while True:
        response = (
            conn.files()
            .list(
                q=f"name contains {_quote(pattern)} and {_quote(folder_id)} in parents ",
                fields="nextPageToken, files(name,id, mimeType,webContentLink,size)",
                pageToken=page_token,
            )
            .execute()
        )
        yield from (f["webContentLink"] for f in response["files"])
        page_token = response.get("nextPageToken", None)
        if page_token is None:
            return


class GdriveLocation(BaseFileLocation):
    """Handler for Google Drive operators."""

    location_type = FileLocation.GOOGLE_DRIVE

    @property
    def hook(self) -> GoogleDriveHook:
        if self.conn_id:
            return GoogleDriveHook(gcp_conn_id=self.conn_id)
        return GoogleDriveHook()

    @property
    def transport_params(self) -> dict:
        """Get Google Drive client."""
        return {"client": self.hook.get_conn()}

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        parsed_url = urllib.parse.urlsplit(self.path)
        return f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return urllib.parse.urlsplit(self.path).path

    def _get_rel_path_parts(self) -> Sequence[str]:
        p = urllib.parse.urlsplit(self.path).path.lstrip("/")
        return pathlib.PurePosixPath(p).parts

    def _get_conn(self) -> Any:
        return self.hook.get_conn()

    @property
    def paths(self) -> list[str]:
        path_parts = self._get_rel_path_parts()
        if not path_parts:
            return []

        conn = self._get_conn()
        url = urllib.parse.urlsplit(self.path)
        result_paths = []
        try:
            folder_id = _find_item_id(url.netloc + url.path, conn=conn)
        except FileNotFoundError:
            raise FileNotFoundError(self.path) from None
        for name in _find_contains(folder_id, path_parts[-1], conn=conn):
            result_paths.append(name)
        return result_paths

    @property
    def size(self) -> int:
        path_parts = self._get_rel_path_parts()
        conn = self._get_conn()
        url = urllib.parse.urlsplit(self.path)
        try:
            folder_id = _find_item_id(url.netloc + url.path, conn=conn)
        except FileNotFoundError:
            raise FileNotFoundError(self.path) from None

        response = (
            conn.files()
            .list(
                q=f"name contains {_quote(path_parts[-1])} and {_quote(folder_id)} in parents ",
                fields="files(name,id,size)",
            )
            .execute()
        )
        value: int = 0
        try:
            for f in response["files"]:
                value = int(f["size"])
        except KeyError:
            raise IsADirectoryError(self.path) from None
        return value



import tempfile

from airflow.configuration import conf

from astro.constants import DEFAULT_SCHEMA

# Section name for astro SDK configs in airflow.cfg
SECTION_KEY = "astro_sdk"
# Bigquery list of all the valid locations: https://cloud.google.com/bigquery/docs/locations
DEFAULT_BIGQUERY_SCHEMA_LOCATION = "us"

SCHEMA = conf.get(SECTION_KEY, "sql_schema", fallback=DEFAULT_SCHEMA)
POSTGRES_SCHEMA = conf.get(SECTION_KEY, "postgres_default_schema", fallback=SCHEMA)
BIGQUERY_SCHEMA = conf.get(SECTION_KEY, "bigquery_default_schema", fallback=SCHEMA)
SNOWFLAKE_SCHEMA = conf.get(SECTION_KEY, "snowflake_default_schema", fallback=SCHEMA)
REDSHIFT_SCHEMA = conf.get(SECTION_KEY, "redshift_default_schema", fallback=SCHEMA)

BIGQUERY_SCHEMA_LOCATION = conf.get(
    SECTION_KEY, "bigquery_dataset_location", fallback=DEFAULT_BIGQUERY_SCHEMA_LOCATION
)

LOAD_FILE_ENABLE_NATIVE_FALLBACK = conf.get(SECTION_KEY, "load_file_enable_native_fallback", fallback=False)

DATAFRAME_STORAGE_CONN_ID = conf.get(SECTION_KEY, "xcom_storage_conn_id", fallback=None)
DATAFRAME_STORAGE_URL = conf.get(SECTION_KEY, "xcom_storage_url", fallback=tempfile.gettempdir())
STORE_DATA_LOCAL_DEV = conf.get(SECTION_KEY, "store_data_local_dev", fallback=False)

# Size is in KB - Max memory usage for Dataframe to be stored in the Airflow metadata DB
# If dataframe size is more than that, it will be stored in an Object Store defined by
# DATAFRAME_STORAGE_CONN_ID & DATAFRAME_STORAGE_URL above
MAX_DATAFRAME_MEMORY_FOR_XCOM_DB = conf.getint(SECTION_KEY, "max_dataframe_mem_for_xcom_db", fallback=100)

OPENLINEAGE_EMIT_TEMP_TABLE_EVENT = conf.getboolean(
    SECTION_KEY, "openlineage_emit_temp_table_event", fallback=True
)
XCOM_BACKEND = conf.get("core", "xcom_backend")
IS_CUSTOM_XCOM_BACKEND = XCOM_BACKEND not in [
    "airflow.models.xcom.BaseXCom",
    "astro.custom_backend.astro_custom_backend.AstroCustomXcomBackend",
]
# We are not defining a fallback key on purpose. S3 Snowflake stages can also
# be created without a storage integration, by using the Airflow AWS connection
# properties.
SNOWFLAKE_STORAGE_INTEGRATION_AMAZON = conf.get(
    section=SECTION_KEY, key="snowflake_storage_integration_amazon", fallback=None
)

SNOWFLAKE_STORAGE_INTEGRATION_GOOGLE = conf.get(
    section=SECTION_KEY, key="snowflake_storage_integration_google", fallback=None
)

#: How many file rows should be loaded to infer the table columns types
LOAD_TABLE_AUTODETECT_ROWS_COUNT = conf.getint(
    section=SECTION_KEY, key="load_table_autodetect_rows_count", fallback=1000
)


#: Reduce responses sizes returned by aql.run_raw_sql to avoid trashing the Airflow DB if the BaseXCom is used.
RAW_SQL_MAX_RESPONSE_SIZE = conf.getint(section=SECTION_KEY, key="run_raw_sql_response_size", fallback=-1)

# Temp changes
# Should Astro SDK automatically add inlets/outlets to take advantage of Airflow 2.4 Data-aware scheduling
AUTO_ADD_INLETS_OUTLETS = conf.getboolean(SECTION_KEY, "auto_add_inlets_outlets", fallback=True)



from typing import cast

from airflow.decorators.base import get_unique_task_id


def get_task_id(prefix: str, path: str) -> str:
    """Generate unique tasks id based on the path.

    :param prefix: prefix string
    :param path: file path.
    """
    task_id = f"{prefix}_{path.rsplit('/', 1)[-1].replace('.', '_')}"
    return cast(str, get_unique_task_id(task_id))



class NonExistentTableException(Exception):
    """Raised if an operation expected a SQL table to exist, but it does not exist"""


class IllegalLoadToDatabaseException(Exception):
    def __init__(self):  # pragma: no cover
        self.message = (
            "Failing this task because you do not have a custom xcom backend set up. If you use "
            "the default XCOM backend to store large dataframes, this can significantly degrade "
            "Airflow DB performance. Please set up a custom XCOM backend (info here "
            "https://docs.astronomer.io/learn/custom-xcom-backends) or set the environment "
            "variable AIRFLOW__ASTRO_SDK__DATAFRAME_ALLOW_UNSAFE_STORAGE to true if you wish to proceed while "
            "knowing the risks. "
        )
        super().__init__(self.message)


class DatabaseCustomError(ValueError, AttributeError):
    """
    Inappropriate argument value (of correct type) or attribute
    not found while running query. while running query
    """



from airflow.exceptions import AirflowException

try:
    # Airflow >= 2.3
    from airflow.models.abstractoperator import AbstractOperator
except ImportError:  # pragma: no cover
    # Airflow < 2.3
    from airflow.models.baseoperator import BaseOperator as AbstractOperator

from airflow.models.xcom_arg import XComArg


class UpstreamTaskMixin:
    def __init__(self, **kwargs):
        upstream_tasks = kwargs.pop("upstream_tasks", [])

        super().__init__(**kwargs)

        for task in upstream_tasks:
            if isinstance(task, XComArg):
                self.set_upstream(task.operator)
            elif isinstance(task, AbstractOperator):
                self.set_upstream(task)
            else:
                raise AirflowException(
                    "Cannot upstream a non-task, please only use XcomArg or operators for this" " parameter"
                )



from __future__ import annotations

import socket

from airflow.providers.sqlite.hooks.sqlite import SqliteHook
from sqlalchemy import MetaData as SqlaMetaData, create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.sql.schema import Table as SqlaTable

from astro.constants import MergeConflictStrategy
from astro.databases.base import BaseDatabase
from astro.table import BaseTable, Metadata

DEFAULT_CONN_ID = SqliteHook.default_conn_name


class SqliteDatabase(BaseDatabase):
    """
    Handle interactions with Sqlite databases. If this class is successful, we should not have any Sqlite-specific
    logic in other parts of our code-base.
    """

    def __init__(self, conn_id: str = DEFAULT_CONN_ID, table: BaseTable | None = None):
        super().__init__(conn_id)
        self.table = table

    @property
    def sql_type(self) -> str:
        return "sqlite"

    @property
    def hook(self) -> SqliteHook:
        """Retrieve Airflow hook to interface with the Sqlite database."""
        return SqliteHook(sqlite_conn_id=self.conn_id)

    @property
    def sqlalchemy_engine(self) -> Engine:
        """Return SQAlchemy engine."""
        # Airflow uses sqlite3 library and not SqlAlchemy for SqliteHook
        # and it only uses the hostname directly.
        airflow_conn = self.hook.get_connection(self.conn_id)
        return create_engine(f"sqlite:///{airflow_conn.host}")

    @property
    def default_metadata(self) -> Metadata:
        """Since Sqlite does not use Metadata, we return an empty Metadata instances."""
        return Metadata()

    # ---------------------------------------------------------
    # Table metadata
    # ---------------------------------------------------------
    @staticmethod
    def get_table_qualified_name(table: BaseTable) -> str:
        """
        Return the table qualified name.

        :param table: The table we want to retrieve the qualified name for.
        """
        return str(table.name)

    def populate_table_metadata(self, table: BaseTable) -> BaseTable:
        """
        Since SQLite does not have a concept of databases or schemas, we just return the table as is,
        without any modifications.
        """
        table.conn_id = table.conn_id or self.conn_id
        return table

    def create_schema_if_needed(self, schema: str | None, location: str | None = None) -> None:
        """
        Since SQLite does not have schemas, we do not need to set a schema here.
        """

    def schema_exists(self, schema: str) -> bool:  # skipcq PYL-W0613,PYL-R0201
        """
        Check if a schema exists. We return false for sqlite since sqlite does not have schemas
        """
        return False

    @staticmethod
    def get_merge_initialization_query(parameters: tuple) -> str:
        """
        Handles database-specific logic to handle index for Sqlite.
        """
        return "CREATE UNIQUE INDEX merge_index ON {{table}}(%s)" % ",".join(parameters)  # skipcq PYL-C0209

    def merge_table(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        if_conflicts: MergeConflictStrategy = "exception",
    ) -> None:
        """
        Merge the source table rows into a destination table.
        The argument `if_conflicts` allows the user to define how to handle conflicts.

        :param source_table: Contains the rows to be merged to the target_table
        :param target_table: Contains the destination table in which the rows will be merged
        :param source_to_target_columns_map: Dict of target_table columns names to source_table columns names
        :param target_conflict_columns: List of cols where we expect to have a conflict while combining
        :param if_conflicts: The strategy to be applied if there are conflicts.
        """
        statement = "INSERT INTO {main_table} ({target_columns}) SELECT {append_columns} FROM {source_table} Where true"
        if if_conflicts == "ignore":
            statement += " ON CONFLICT ({merge_keys}) DO NOTHING"
        elif if_conflicts == "update":
            statement += " ON CONFLICT ({merge_keys}) DO UPDATE SET {update_statements}"

        append_column_names = list(source_to_target_columns_map.keys())
        target_column_names = list(source_to_target_columns_map.values())
        update_statements = [f"{col_name}=EXCLUDED.{col_name}" for col_name in target_column_names]

        query = statement.format(
            target_columns=",".join(target_column_names),
            main_table=target_table.name,
            append_columns=",".join(append_column_names),
            source_table=source_table.name,
            update_statements=",".join(update_statements),
            merge_keys=",".join(list(target_conflict_columns)),
        )

        self.run_sql(sql=query)

    def get_sqla_table(self, table: BaseTable) -> SqlaTable:
        """
        Return SQLAlchemy table instance

        :param table: Astro Table to be converted to SQLAlchemy table instance
        """
        return SqlaTable(table.name, SqlaMetaData(), autoload_with=self.sqlalchemy_engine)

    def openlineage_dataset_name(self, table: BaseTable) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: /tmp/local.db.table_name
        """
        conn = self.hook.get_connection(self.conn_id)
        return f"{conn.host}.{table.name}"

    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: file://127.0.0.1:22
        """
        conn = self.hook.get_connection(self.conn_id)
        port = conn.port or 22
        return f"file://{socket.gethostbyname(socket.gethostname())}:{port}"

    def openlineage_dataset_uri(self, table: BaseTable) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return f"{self.openlineage_dataset_namespace()}{self.openlineage_dataset_name(table=table)}"



from __future__ import annotations

from urllib.parse import urlparse, urlunparse

from airflow.providers.google.cloud.hooks.gcs import GCSHook

from astro.constants import FileLocation
from astro.files.locations.base import BaseFileLocation


class GCSLocation(BaseFileLocation):
    """Handler GS object store operations"""

    location_type = FileLocation.GS

    @property
    def hook(self) -> GCSHook:
        return GCSHook(gcp_conn_id=self.conn_id) if self.conn_id else GCSHook()

    @property
    def transport_params(self) -> dict:
        """get GCS credentials for storage"""
        client = self.hook.get_conn()
        return {"client": client}

    @property
    def paths(self) -> list[str]:
        """Resolve GS file paths with prefix"""
        url = urlparse(self.path)
        bucket_name = url.netloc
        prefix = url.path[1:]
        prefixes = self.hook.list(bucket_name=bucket_name, prefix=prefix)
        paths = [urlunparse((url.scheme, url.netloc, keys, "", "", "")) for keys in prefixes]
        return paths

    @property
    def size(self) -> int:
        """Return file size for GCS location"""
        url = urlparse(self.path)
        bucket_name = url.netloc
        object_name = url.path
        if object_name.startswith("/"):
            object_name = object_name[1:]
        return int(self.hook.get_size(bucket_name=bucket_name, object_name=object_name))

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        parsed_url = urlparse(self.path)
        return f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return urlparse(self.path).path



from __future__ import annotations

import warnings
from typing import Any

import pandas as pd
from airflow.models.xcom_arg import XComArg

from astro.constants import ExportExistsStrategy
from astro.files import File
from astro.sql.operators.export_table_to_file import ExportTableToFileOperator, export_table_to_file
from astro.table import BaseTable


class ExportFileOperator(ExportTableToFileOperator):
    """Write SQL table to csv/parquet on local/S3/GCS.

    :param input_data: Table to convert to file
    :param output_file: File object containing the path to the file and connection id.
    :param if_exists: Overwrite file if exists. Default False.
    """

    template_fields = ("input_data", "output_file")

    def __init__(
        self,
        input_data: BaseTable | pd.DataFrame,
        output_file: File,
        if_exists: ExportExistsStrategy = "exception",
        **kwargs,
    ) -> None:
        super().__init__(input_data=input_data, output_file=output_file, if_exists=if_exists, **kwargs)
        warnings.warn(
            """This class is deprecated.
            Please use `astro.sql.operators.export_table_to_file.ExportTableToFileOperator`.
            And, will be removed in astro-sdk-python>=2.0.0.""",
            DeprecationWarning,
            stacklevel=2,
        )


def export_file(
    input_data: BaseTable | pd.DataFrame,
    output_file: File,
    if_exists: ExportExistsStrategy = "exception",
    task_id: str | None = None,
    **kwargs: Any,
) -> XComArg:
    """Convert ExportFileOperator into a function. Returns XComArg.

    Returns an XComArg object of type File which matches the output_file parameter.

    This will allow users to perform further actions with the exported file.

    e.g.:

    .. code-block:: python

      with sample_dag:
          table = aql.load_file(input_file=File(path=data_path), output_table=test_table)
          exported_file = aql.export_file(
              input_data=table,
              output_file=File(path="/tmp/saved_df.csv"),
              if_exists="replace",
          )
          res_df = aql.load_file(input_file=exported_file)


    :param output_file: Path and conn_id
    :param input_data: Input table / dataframe
    :param if_exists: Overwrite file if exists. Default "exception"
    :param task_id: task id, optional
    """

    warnings.warn(
        """This decorator is deprecated.
        Please use `astro.sql.operators.export_table_to_file.export_table_to_file`.
        And, will be removed in astro-sdk-python>=2.0.0.""",
        DeprecationWarning,
        stacklevel=2,
    )

    return export_table_to_file(
        input_data=input_data, output_file=output_file, if_exists=if_exists, task_id=task_id, **kwargs
    )
